<?php

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Export;

use Chill\ActivityBundle\Repository\ActivityRepository;
use Chill\MainBundle\Export\ExportInterface;
use Chill\MainBundle\Export\FormatterInterface;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Role\Role;
use Doctrine\ORM\Query;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;

/**
 * This export allow to compute stats on activity duration.
 *
 * The desired stat must be given in constructor.
 */
class StatActivityDuration implements ExportInterface
{
    private ActivityRepository $activityRepository;

    public const SUM = 'sum';

    /**
     * The action for this report.
     */
    protected string $action;

    /**
     * @param string $action the stat to perform
     */
    public function __construct(
        ActivityRepository $activityRepository,
        string $action = 'sum'
    ) {
        $this->action = $action;
        $this->activityRepository = $activityRepository;
    }

    public function buildForm(FormBuilderInterface $builder)
    {

    }

    public function getDescription()
    {
        if ($this->action === self::SUM) {
            return 'Sum activities duration by various parameters.';
        }
    }

    public function getTitle()
    {
        if ($this->action === self::SUM) {
            return 'Sum activity duration';
        }

    }

    public function getType()
    {
        return 'activity';
    }

    public function initiateQuery(array $requiredModifiers,  array $acl, array $data = [])
    {
        $centers = array_map(
            static fn(array $el): string => $el['center'],
            $acl
        );

        $qb = $this->activityRepository->createQueryBuilder('activity');

        $select = null;

        if ($this->action === self::SUM) {
            $select = 'SUM(activity.durationTime) AS export_stat_activity';
        }

        return $qb->select($select)
            ->join('activity.person',  'person')
            ->join('person.center', 'center')
            ->where($qb->expr()->in('center', ':centers'))
            ->setParameter(':centers', $centers);
    }

    public function supportsModifiers()
    {
        return ['person', 'activity'];
    }

    public function requiredRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_TABULAR];
    }

    public function getLabels($key, array $values, $data)
    {
        if ($key !== 'export_stat_activity') {
            throw new \LogicException(sprintf('The key %s is not used by this export', $key));
        }

        $header = $this->action === self::SUM ? 'Sum of activities duration' : false;

        return static fn(string $value) => $value === '_header' ?  $header : $value;
    }

    public function getQueryKeys($data)
    {
        return ['export_stat_activity'];
    }

    public function getResult($qb, $data)
    {
        return $qb->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

}
