<?php

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Filter;

use Chill\MainBundle\Export\FilterInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormEvent;
use Symfony\Component\Form\FormEvents;
use Symfony\Component\Form\Extension\Core\Type\DateType;
use Symfony\Component\Form\FormError;
use Chill\MainBundle\Form\Type\Export\FilterType;
use Doctrine\ORM\Query\Expr;
use Symfony\Component\Translation\TranslatorInterface;

class ActivityDateFilter implements FilterInterface
{
    protected TranslatorInterface $translator;

    function __construct(TranslatorInterface $translator)
    {
        $this->translator = $translator;
    }

    public function addRole()
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $where = $qb->getDQLPart('where');
        $clause = $qb->expr()->between('activity.date', ':date_from',
            ':date_to');

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }

        $qb->add('where', $where);
        $qb->setParameter('date_from', $data['date_from']);
        $qb->setParameter('date_to', $data['date_to']);
    }

    public function applyOn()
    {
        return 'activity';
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add(
            'date_from',
            DateType::class,
            [
                'label' => 'Activities after this date',
                'data'  => new \DateTime(),
                'attr'  => ['class' => 'datepicker'],
                'widget'=> 'single_text',
                'format' => 'dd-MM-yyyy',
            ]
        );

        $builder->add(
            'date_to',
            DateType::class,
            [
                'label' => 'Activities before this date',
                'data'  => new \DateTime(),
                'attr'  => ['class' => 'datepicker'],
                'widget'=> 'single_text',
                'format' => 'dd-MM-yyyy',
            ]
        );

        $builder->addEventListener(FormEvents::POST_SUBMIT, function(FormEvent $event) {
            /* @var $filterForm \Symfony\Component\Form\FormInterface */
            $filterForm = $event->getForm()->getParent();
            $enabled = $filterForm->get(FilterType::ENABLED_FIELD)->getData();

            if ($enabled === true) {
                // if the filter is enabled, add some validation
                $form      = $event->getForm();
                $date_from = $form->get('date_from')->getData();
                $date_to   = $form->get('date_to')->getData();

                // check that fields are not empty
                if ($date_from === null) {
                    $form->get('date_from')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')));
                }
                if ($date_to   === null) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')));
                }

                // check that date_from is before date_to
                if (
                    ($date_from !== null && $date_to !== null)
                    &&
                    $date_from >= $date_to
                ) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This date should be after '
                            . 'the date given in "Implied in an activity after '
                            . 'this date" field')));
                }
            }
        });
    }

    public function describeAction($data, $format = 'string')
    {
        return [
            'Filtered by date of activity: only between %date_from% and %date_to%',
            [
                '%date_from%' => $data['date_from']->format('d-m-Y'),
                '%date_to%' => $data['date_to']->format('d-m-Y')
            ]
        ];
    }

    public function getTitle()
    {
        return 'Filtered by date activity';
    }

}
