<?php

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Filter;

use Chill\ActivityBundle\Repository\ActivityReasonRepository;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Chill\ActivityBundle\Entity\ActivityReason;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\Query\Expr;
use Symfony\Component\Security\Core\Role\Role;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Doctrine\ORM\Query\Expr\Join;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Chill\MainBundle\Export\ExportElementValidatedInterface;

class ActivityReasonFilter implements FilterInterface, ExportElementValidatedInterface
{
    protected TranslatableStringHelperInterface $translatableStringHelper;

    protected ActivityReasonRepository $activityReasonRepository;

    public function __construct(
        TranslatableStringHelper $helper,
        ActivityReasonRepository $activityReasonRepository
    ) {
        $this->translatableStringHelper = $helper;
        $this->activityReasonRepository = $activityReasonRepository;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $where = $qb->getDQLPart('where');
        $join  = $qb->getDQLPart('join');
        $clause = $qb->expr()->in('reasons', ':selected_activity_reasons');
        //dump($join);
        // add a join to reasons only if needed
        if (
            (array_key_exists('activity', $join)
            &&
            !$this->checkJoinAlreadyDefined($join['activity'], 'reasons')
            )
            ||
            (! array_key_exists('activity', $join))
        ) {
            $qb->add(
                    'join',
                    array('activity' => new Join(Join::INNER_JOIN, 'activity.reasons', 'reasons')),
                    true
                    );
        }

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }

        $qb->add('where', $where);
        $qb->setParameter('selected_activity_reasons', $data['reasons']);
    }

    /**
     * Check if a join between Activity and Reason is already defined
     *
     * @param Join[] $joins
     * @return boolean
     */
    private function checkJoinAlreadyDefined(array $joins, $alias): bool
    {
        foreach ($joins as $join) {
            if ($join->getAlias() === $alias) {
                return true;
            }
        }

        return false;
    }

    public function applyOn()
    {
        return 'activity';
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('reasons', EntityType::class, [
            'class' => ActivityReason::class,
            'choice_label' => fn(ActivityReason $reason) => $this->translatableStringHelper->localize($reason->getName()),
            'group_by' => fn(ActivityReason $reason) => $this->translatableStringHelper->localize($reason->getCategory()->getName()),
            'multiple' => true,
            'expanded' => false
        ]);
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        if ($data['reasons'] === null || count($data['reasons']) === 0) {
            $context
                ->buildViolation('At least one reason must be chosen')
                ->addViolation();
        }
    }

    public function getTitle()
    {
        return 'Filter by reason';
    }

    public function addRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }

    public function describeAction($data, $format = 'string')
    {
        // collect all the reasons'name used in this filter in one array
        $reasonsNames = array_map(
            fn(ActivityReason $r): string => '"' . $this->translatableStringHelper->localize($r->getName()) . '"',
            $this->activityReasonRepository->findBy(array('id' => $data['reasons']->toArray()))
        );

        return [
            'Filtered by reasons: only %list%',
            [
                '%list%' => implode(", ", $reasonsNames),
            ]
        ];
    }
}
