<?php

/*
 * Copyright (C) 2015 Julien Fastré <julien.fastre@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\ActivityBundle\Security\Authorization;

use Chill\MainBundle\Security\Authorization\VoterHelperFactoryInterface;
use Chill\MainBundle\Security\Authorization\VoterHelperInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;

use Chill\MainBundle\Security\Authorization\AbstractChillVoter;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\MainBundle\Security\ProvideRoleHierarchyInterface;
use Chill\MainBundle\Entity\User;
use Chill\ActivityBundle\Entity\Activity;
use Chill\PersonBundle\Entity\Person;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Security\Core\Security;

/**
 * Voter for Activity class
 */
class ActivityVoter extends AbstractChillVoter implements ProvideRoleHierarchyInterface
{
    const CREATE = 'CHILL_ACTIVITY_CREATE';
    const SEE    = 'CHILL_ACTIVITY_SEE';
    const SEE_DETAILS = 'CHILL_ACTIVITY_SEE_DETAILS';
    const UPDATE = 'CHILL_ACTIVITY_UPDATE';
    const DELETE = 'CHILL_ACTIVITY_DELETE';
    const FULL = 'CHILL_ACTIVITY_FULL';

    private const ALL = [
        self::CREATE,
        self::SEE,
        self::UPDATE,
        self::DELETE,
        self::SEE_DETAILS,
        self::FULL
    ];

    protected VoterHelperInterface $voterHelper;

    protected Security $security;

    public function __construct(
        Security $security,
        VoterHelperFactoryInterface $voterHelperFactory
    ) {
        $this->security = $security;
        $this->voterHelper = $voterHelperFactory->generate(self::class)
            ->addCheckFor(Person::class, [self::SEE, self::CREATE])
            ->addCheckFor(AccompanyingPeriod::class, [self::SEE, self::CREATE])
            ->addCheckFor(Activity::class, self::ALL)
            ->build();
    }


    protected function supports($attribute, $subject)
    {
        return $this->voterHelper->supports($attribute, $subject);
    }

    protected function voteOnAttribute($attribute, $subject, TokenInterface $token)
    {
        if (!$token->getUser() instanceof User) {
            return false;
        }

        if ($subject instanceof Activity) {
            if ($subject->getPerson() instanceof Person) {
                // the context is person: we must have the right to see the person
                if (!$this->security->isGranted(PersonVoter::SEE, $subject->getPerson())) {
                    return false;
                }
            } elseif ($subject->getAccompanyingPeriod() instanceof AccompanyingPeriod) {
                if (!$this->security->isGranted(AccompanyingPeriodVoter::SEE, $subject->getAccompanyingPeriod())) {
                    return false;
                }
            } else {
                throw new \RuntimeException("could not determine context of activity");
            }
        }

        return $this->voterHelper->voteOnAttribute($attribute, $subject, $token);
    }


    public function getRoles()
    {
        return self::ALL;
    }

    public function getRolesWithoutScope()
    {
        return [];
    }


    public function getRolesWithHierarchy()
    {
        return [ 'Activity' => $this->getRoles() ];
    }

}
