<?php

namespace Chill\DocGeneratorBundle\Serializer\Normalizer;

use Chill\DocGeneratorBundle\Serializer\Helper\NormalizeNullValueHelper;
use Symfony\Component\PropertyAccess\PropertyAccess;
use Symfony\Component\PropertyAccess\PropertyAccessor;
use Symfony\Component\Serializer\Exception\ExceptionInterface;
use Symfony\Component\Serializer\Exception\LogicException;
use Symfony\Component\Serializer\Mapping\AttributeMetadata;
use Symfony\Component\Serializer\Mapping\ClassMetadata;
use Symfony\Component\Serializer\Mapping\Factory\ClassMetadataFactoryInterface;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareTrait;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

class DocGenObjectNormalizer implements NormalizerInterface, NormalizerAwareInterface
{
    use NormalizerAwareTrait;

    private ClassMetadataFactoryInterface $classMetadataFactory;
    private PropertyAccessor $propertyAccess;

    public function __construct(ClassMetadataFactoryInterface $classMetadataFactory)
    {
        $this->classMetadataFactory = $classMetadataFactory;
        $this->propertyAccess = PropertyAccess::createPropertyAccessor();
    }

    /**
     * @inheritDoc
     */
    public function normalize($object, string $format = null, array $context = [])
    {
        $classMetadataKey = $object ?? $context['docgen:expects'];

        if (!$this->classMetadataFactory->hasMetadataFor($classMetadataKey)) {
            throw new LogicException(sprintf("This object does not have metadata: %s. Add groups on this entity to allow to serialize with the format %s and groups %s", is_object($object) ? get_class($object) : $context['docgen:expects'], $format, \implode(', ', $context['groups'])));
        }

        $metadata = $this->classMetadataFactory->getMetadataFor($classMetadataKey);
        $expectedGroups = \array_key_exists(AbstractNormalizer::GROUPS, $context) ?
            \is_array($context[AbstractNormalizer::GROUPS]) ? $context[AbstractNormalizer::GROUPS] : [$context[AbstractNormalizer::GROUPS]]
            : [];
        $attributes = \array_filter(
            $metadata->getAttributesMetadata(),
            function (AttributeMetadata $a) use ($expectedGroups) {
                foreach ($a->getGroups() as $g) {
                    if (\in_array($g, $expectedGroups, true)) {
                        return true;
                    }
                }

                return false;
            });

        if (null === $object) {
            return $this->normalizeNullData($format, $context, $metadata, $attributes);
        }

        return $this->normalizeObject($object, $format, $context, $expectedGroups, $metadata, $attributes);
    }

    /**
     * @param string $format
     * @param array $context
     * @param array $expectedGroups
     * @param ClassMetadata $metadata
     * @param array|AttributeMetadata[] $attributes
     */
    private function normalizeNullData(string $format, array $context, ClassMetadata $metadata, array $attributes): array
    {
        $keys = [];

        foreach ($attributes as $attribute) {
            $key = $attribute->getSerializedName() ?? $attribute->getName();
            $keys[$key] = $this->getExpectedType($attribute, $metadata->getReflectionClass());
        }

        $normalizer = new NormalizeNullValueHelper($this->normalizer);

        return $normalizer->normalize($keys, $format, $context);
    }

    /**
     * @param $object
     * @param $format
     * @param array $context
     * @param array $expectedGroups
     * @param ClassMetadata $metadata
     * @param array|AttributeMetadata[] $attributes
     * @return array
     * @throws ExceptionInterface
     */
    private function normalizeObject($object, $format, array $context, array $expectedGroups, ClassMetadata $metadata, array $attributes)
    {
        $data = [];
        $reflection = $metadata->getReflectionClass();

        foreach ($attributes as $attribute) {
            /** @var AttributeMetadata $attribute */
            $value = $this->propertyAccess->getValue($object, $attribute->getName());
            $key = $attribute->getSerializedName() ?? $attribute->getName();

            if (is_object($value)) {
                $data[$key] =
                    $this->normalizer->normalize($value, $format, \array_merge(
                        $context, $attribute->getNormalizationContextForGroups($expectedGroups)
                    ));
            } elseif (null === $value) {
                $data[$key] = $this->normalizeNullOutputValue($format, $context, $attribute, $reflection);
            } else {
                $data[$key] = (string) $value;
            }
        }

        return $data;
    }

    private function getExpectedType(AttributeMetadata $attribute, \ReflectionClass $reflection): string
    {
        // we have to get the expected content
        if ($reflection->hasProperty($attribute->getName())) {
            $type = $reflection->getProperty($attribute->getName())->getType();
        } elseif ($reflection->hasMethod($attribute->getName())) {
            $type = $reflection->getMethod($attribute->getName())->getReturnType();
        } else {
            throw new \LogicException(sprintf(
                "Could not determine how the content is determined for the attribute %s. Add attribute property only on property or method", $attribute->getName()
            ));
        }

        if (null === $type) {
            throw new \LogicException(sprintf(
                "Could not determine the type for this attribute: %s. Add a return type to the method or property declaration", $attribute->getName()
            ));
        }

        return $type->getName();
    }

    /**
     */
    private function normalizeNullOutputValue($format, array $context, AttributeMetadata $attribute, \ReflectionClass $reflection)
    {
        $type = $this->getExpectedType($attribute, $reflection);

        switch ($type) {
            case 'array':
            case 'bool':
            case 'double':
            case 'float':
            case 'int':
            case 'resource':
            case 'string':
                return '';
            default:
                return $this->normalizer->normalize(
                    null,
                    $format,
                    \array_merge(
                        $context,
                        ['docgen:expects' => $type]
                    )
                );
        }
    }

    /**
     * @inheritDoc
     */
    public function supportsNormalization($data, string $format = null): bool
    {
        return $format === 'docgen' && (is_object($data) || null === $data);
    }
}
