<?php

namespace Chill\DocStoreBundle\Controller;

use Chill\DocStoreBundle\Entity\AccompanyingCourseDocument;
use Chill\DocStoreBundle\Form\AccompanyingCourseDocumentType;
use Chill\DocStoreBundle\Security\Authorization\AccompanyingCourseDocumentVoter;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Privacy\PrivacyEvent;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Contracts\Translation\TranslatorInterface;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/{_locale}/parcours/{course}/document")
 */
class DocumentAccompanyingCourseController extends AbstractController
{

    /**
     *
     * @var TranslatorInterface
     */
    protected $translator;

    /**
     * @var EventDispatcherInterface
     */
    protected $eventDispatcher;

    /**
     * @var AuthorizationHelper
     */
    protected $authorizationHelper;

    /**
     * DocumentAccompanyingCourseController constructor.

     * @param TranslatorInterface $translator
     * @param EventDispatcherInterface $eventDispatcher
     * @param AuthorizationHelper $authorizationHelper
     */
    public function __construct(
        TranslatorInterface $translator,
        EventDispatcherInterface $eventDispatcher,
        AuthorizationHelper $authorizationHelper
    ) {
        $this->translator = $translator;
        $this->eventDispatcher = $eventDispatcher;
        $this->authorizationHelper = $authorizationHelper;
    }

    /**
     * @Route("/", name="accompanying_course_document_index", methods="GET")
     */
    public function index(AccompanyingPeriod $course): Response
    {
        $em = $this->getDoctrine()->getManager();

        if ($course === NULL) {
            throw $this->createNotFoundException('Accompanying period not found');
        }

        $this->denyAccessUnlessGranted(AccompanyingCourseDocumentVoter::SEE, $course);

        $documents = $em
            ->getRepository("ChillDocStoreBundle:AccompanyingCourseDocument")
            ->findBy(
                ['course' => $course],
                ['date' => 'DESC']
            );

        return $this->render(
            'ChillDocStoreBundle:AccompanyingCourseDocument:index.html.twig',
            [
                'documents' => $documents,
                'accompanyingCourse' => $course
            ]);
    }

    /**
     * @Route("/new", name="accompanying_course_document_new", methods="GET|POST")
     */
    public function new(Request $request, AccompanyingPeriod $course): Response
    {
        if ($course === NULL) {
            throw $this->createNotFoundException('Accompanying period not found');
        }

        $document = new AccompanyingCourseDocument();
        $document->setUser($this->getUser());
        $document->setCourse($course);
        $document->setDate(new \DateTime('Now'));

        $this->denyAccessUnlessGranted(AccompanyingCourseDocumentVoter::CREATE, $document);

        $form = $this->createForm(AccompanyingCourseDocumentType::class, $document);
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $this->denyAccessUnlessGranted(
                'CHILL_ACCOMPANYING_COURSE_DOCUMENT_CREATE', $document,
                'creation of this activity not allowed');

            $em = $this->getDoctrine()->getManager();
            $em->persist($document);
            $em->flush();

            $this->addFlash('success', $this->translator->trans("The document is successfully registered"));

            return $this->redirectToRoute('accompanying_course_document_index', ['course' => $course->getId()]);
        } elseif ($form->isSubmitted() and !$form->isValid()) {
            $this->addFlash('error', $this->translator->trans("This form contains errors"));
        }

        return $this->render('ChillDocStoreBundle:AccompanyingCourseDocument:new.html.twig', [
            'document' => $document,
            'form' => $form->createView(),
            'accompanyingCourse' => $course,
        ]);
    }

    /**
     * @Route("/{id}", name="accompanying_course_document_show", methods="GET")
     */
    public function show(AccompanyingPeriod $course, AccompanyingCourseDocument $document): Response
    {
        $this->denyAccessUnlessGranted(AccompanyingCourseDocumentVoter::SEE_DETAILS, $document);

        return $this->render(
            'ChillDocStoreBundle:AccompanyingCourseDocument:show.html.twig',
            ['document' => $document, 'accompanyingCourse' => $course]);
    }

    /**
     * @Route("/{id}/edit", name="accompanying_course_document_edit", methods="GET|POST")
     */
    public function edit(Request $request, AccompanyingPeriod $course, AccompanyingCourseDocument $document): Response
    {
        $this->denyAccessUnlessGranted(AccompanyingCourseDocumentVoter::UPDATE, $document);

        $document->setUser($this->getUser());
        $document->setDate(new \DateTime('Now'));

        $form = $this->createForm(
            AccompanyingCourseDocumentType::class, $document);
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $this->getDoctrine()->getManager()->flush();

            $this->addFlash('success', $this->translator->trans("The document is successfully updated"));

            return $this->redirectToRoute(
                'accompanying_course_document_edit',
                ['id' => $document->getId(), 'course' => $course->getId()]);

        } elseif ($form->isSubmitted() and !$form->isValid()) {
            $this->addFlash('error', $this->translator->trans("This form contains errors"));
        }

        return $this->render(
            'ChillDocStoreBundle:AccompanyingCourseDocument:edit.html.twig',
            [
                'document' => $document,
                'form' => $form->createView(),
                'accompanyingCourse' => $course,
        ]);
    }

    /**
     * @Route("/{id}", name="accompanying_course_document_delete", methods="DELETE")
     */
    public function delete(Request $request, AccompanyingPeriod $course, AccompanyingCourseDocument $document): Response
    {
        $this->denyAccessUnlessGranted(AccompanyingCourseDocumentVoter::DELETE, $document);

        if ($this->isCsrfTokenValid('delete'.$document->getId(), $request->request->get('_token'))) {
            $em = $this->getDoctrine()->getManager();
            $em->remove($document);
            $em->flush();
        }

        return $this->redirectToRoute(
            'accompanying_course_document_index', ['accompanyingCourse' => $course->getId()]);
    }
}
