<?php

namespace Chill\MainBundle\Entity;

use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Symfony\Component\Serializer\Annotation\Groups;
use Chill\MainBundle\Doctrine\Model\Point;
use Chill\ThirdPartyBundle\Entity\ThirdParty;

/**
 * Address
 *
 * @ORM\Entity
 * @ORM\Table(name="chill_main_address")
 * @ORM\HasLifecycleCallbacks()
 */
class Address
{
    /**
     * @var integer
     *
     * @ORM\Id
     * @ORM\Column(name="id", type="integer")
     * @ORM\GeneratedValue(strategy="AUTO")
     * @Groups({"write"})
     */
    private $id;

    /**
     * @var string
     *
     * @ORM\Column(type="string", length=255)
     * @Groups({"write"})
     */
    private $street = '';

    /**
     * @var string
     *
     * @ORM\Column(type="string", length=255)
     * @Groups({"write"})
     */
    private $streetNumber = '';

    /**
     * @var PostalCode
     *
     * @ORM\ManyToOne(targetEntity="Chill\MainBundle\Entity\PostalCode")
     * @Groups({"write"})
     */
    private $postcode;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=16, nullable=true)
     * @Groups({"write"})
     */
    private $floor;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=16, nullable=true)
     * @Groups({"write"})
     */
    private $corridor;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=16, nullable=true)
     * @Groups({"write"})
     */
    private $steps;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=255, nullable=true)
     * @Groups({"write"})
     */
    private $buildingName;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=16, nullable=true)
     * @Groups({"write"})
     */
    private $flat;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=255, nullable=true)
     * @Groups({"write"})
     */
    private $distribution;

    /**
     * @var string|null
     *
     * @ORM\Column(type="string", length=255, nullable=true)
     * @Groups({"write"})
     */
    private $extra;

    /**
     * Indicates when the address starts validation. Used to build an history
     * of address. By default, the current date.
     *
     * @var \DateTime
     *
     * @ORM\Column(type="date")
     * @Groups({"write"})
     */
    private \DateTime $validFrom;

    /**
     * Indicates when the address ends. Used to build an history
     * of address.
     *
     * @var \DateTime|null
     *
     * @ORM\Column(type="date", nullable=true)
     * @Groups({"write"})
     */
    private ?\DateTime $validTo = null;

    /**
     * True if the address is a "no address", aka homeless person, ...
     * @Groups({"write"})
     * @ORM\Column(type="boolean")
     *
     * @var bool
     */
    private $isNoAddress = false;

    /**
     * A geospatial field storing the coordinates of the Address
     *
     * @var Point|null
     *
     * @ORM\Column(type="point", nullable=true)
     * @Groups({"write"})
     */
    private $point;

    /**
     * A ThirdParty reference for person's addresses that are linked to a third party
     *
     * @var ThirdParty|null
     *
     * @ORM\ManyToOne(targetEntity="Chill\ThirdPartyBundle\Entity\ThirdParty")
     * @Groups({"write"})
     * @ORM\JoinColumn(nullable=true, onDelete="SET NULL")
     */
    private $linkedToThirdParty;

    /**
     * A list of metadata, added by customizable fields
     *
     * @var array
     */
    private $customs = [];

    /**
     * @ORM\ManyToOne(targetEntity=AddressReference::class)
     * @Groups({"write"})
     */
    private ?AddressReference $addressReference = null;

    public function __construct()
    {
        $this->validFrom = new \DateTime();
    }

    /**
     * Get id
     *
     * @return integer
     */
    public function getId()
    {
        return $this->id;
    }

    /**
     * Set streetAddress1 (legacy function)
     *
     * @param string $streetAddress1
     *
     * @return Address
     */
    public function setStreetAddress1($streetAddress1)
    {
        $this->street = $streetAddress1 === NULL ? '' : $streetAddress1;

        return $this;
    }

    /**
     * Get streetAddress1 (legacy function)
     *
     * @return string
     */
    public function getStreetAddress1()
    {
        return $this->street;
    }

    /**
     * Set streetAddress2 (legacy function)
     *
     * @param string $streetAddress2
     *
     * @return Address
     */
    public function setStreetAddress2($streetAddress2)
    {
        $this->streetNumber = $streetAddress2 === NULL ? '' : $streetAddress2;

        return $this;
    }

    /**
     * Get streetAddress2 (legacy function)
     *
     * @return string
     */
    public function getStreetAddress2()
    {
        return $this->streetNumber;
    }

    /**
     * Set postcode
     *
     * @param PostalCode $postcode
     *
     * @return Address
     */
    public function setPostcode(PostalCode $postcode = null)
    {
        $this->postcode = $postcode;

        return $this;
    }

    /**
     * Get postcode
     *
     * @return PostalCode
     */
    public function getPostcode()
    {
        return $this->postcode;
    }

    /**
     * @return \DateTime
     */
    public function getValidFrom()
    {
        return $this->validFrom;
    }

    /**
     * @param \DateTime $validFrom
     * @return Address
     */
    public function setValidFrom(\DateTime $validFrom)
    {
        $this->validFrom = $validFrom;
        return $this;
    }

    /**
     * Get IsNoAddress
     *
     * Indicate true if the address is a fake address (homeless, ...)
     *
     * @return bool
     */
    public function getIsNoAddress(): bool
    {
        return $this->isNoAddress;
    }

    /**
     * @return bool
     */
    public function isNoAddress(): bool
    {
        return $this->getIsNoAddress();
    }

    /**
     * Set IsNoAddress
     *
     * Indicate true if the address is a fake address (homeless, ...)
     *
     * @param bool $isNoAddress
     * @return $this
     */
    public function setIsNoAddress(bool $isNoAddress): self
    {
        $this->isNoAddress = $isNoAddress;
        return $this;
    }

    /**
     * Get customs informations in the address
     *
     * @return array
     */
    public function getCustoms(): array
    {
        return $this->customs;
    }

    /**
     * Store custom informations in the address
     *
     * @param array $customs
     * @return $this
     */
    public function setCustoms(array $customs): self
    {
        $this->customs = $customs;

        return $this;
    }

    /**
     * Validate the address.
     *
     * Check that:
     *
     * * if the address is not home address:
     *     * the postal code is present
     *     * the valid from is not null
     *     * the address street 1 is greater than 2
     *
     * @param ExecutionContextInterface $context
     * @param array $payload
     */
    public function validate(ExecutionContextInterface $context, $payload)
    {
        if (!$this->getValidFrom() instanceof \DateTime) {
            $context
                ->buildViolation("address.date-should-be-set")
                ->atPath('validFrom')
                ->addViolation();
        }

        if ($this->isNoAddress()) {
            return;
        }

        if (empty($this->getStreetAddress1())) {
            $context
                ->buildViolation("address.street1-should-be-set")
                ->atPath('streetAddress1')
                ->addViolation();
        }

        if (!$this->getPostcode() instanceof PostalCode) {
            $context
                ->buildViolation("address.postcode-should-be-set")
                ->atPath('postCode')
                ->addViolation();
        }
    }

    /**
     * @param Address $original
     * @return Address
     */
    public static function createFromAddress(Address $original) : Address
    {
        return (new Address())
            ->setAddressReference($original->getAddressReference())
            ->setBuildingName($original->getBuildingName())
            ->setCorridor($original->getCorridor())
            ->setCustoms($original->getCustoms())
            ->setDistribution($original->getDistribution())
            ->setExtra($original->getExtra())
            ->setFlat($original->getFlat())
            ->setFloor($original->getFloor())
            ->setIsNoAddress($original->getIsNoAddress())
            ->setLinkedToThirdParty($original->getLinkedToThirdParty())
            ->setPoint($original->getPoint())
            ->setPostcode($original->getPostcode())
            ->setSteps($original->getSteps())
            ->setStreet($original->getStreet())
            ->setStreetNumber($original->getStreetNumber())
            ->setValidFrom($original->getValidFrom())
            ->setValidTo($original->getValidTo())
            ;
    }

    public static function createFromAddressReference(AddressReference $original): Address
    {
        return (new Address())
            ->setPoint($original->getPoint())
            ->setPostcode($original->getPostcode())
            ->setStreet($original->getStreet())
            ->setStreetNumber($original->getStreetNumber())
            ->setAddressReference($original)
            ;
    }

    public function getStreet(): ?string
    {
        return $this->street;
    }

    public function setStreet(string $street): self
    {
        $this->street = $street;

        return $this;
    }

    public function getStreetNumber(): ?string
    {
        return $this->streetNumber;
    }

    public function setStreetNumber(string $streetNumber): self
    {
        $this->streetNumber = $streetNumber;

        return $this;
    }

    public function getFloor(): ?string
    {
        return $this->floor;
    }

    public function setFloor(?string $floor): self
    {
        $this->floor = $floor;

        return $this;
    }

    public function getCorridor(): ?string
    {
        return $this->corridor;
    }

    public function setCorridor(?string $corridor): self
    {
        $this->corridor = $corridor;

        return $this;
    }

    public function getSteps(): ?string
    {
        return $this->steps;
    }

    public function setSteps(?string $steps): self
    {
        $this->steps = $steps;

        return $this;
    }

    public function getBuildingName(): ?string
    {
        return $this->buildingName;
    }

    public function setBuildingName(?string $buildingName): self
    {
        $this->buildingName = $buildingName;

        return $this;
    }

    public function getFlat(): ?string
    {
        return $this->flat;
    }

    public function setFlat(?string $flat): self
    {
        $this->flat = $flat;

        return $this;
    }

    public function getDistribution(): ?string
    {
        return $this->distribution;
    }

    public function setDistribution(?string $distribution): self
    {
        $this->distribution = $distribution;

        return $this;
    }

    public function getExtra(): ?string
    {
        return $this->extra;
    }

    public function setExtra(?string $extra): self
    {
        $this->extra = $extra;

        return $this;
    }

    public function getValidTo(): ?\DateTimeInterface
    {
        return $this->validTo;
    }

    public function setValidTo(?\DateTimeInterface $validTo = null): self
    {
        $this->validTo = $validTo;

        return $this;
    }

    public function getPoint(): ?Point
    {
        return $this->point;
    }

    public function setPoint(?Point $point): self
    {
        $this->point = $point;

        return $this;
    }

    public function getLinkedToThirdParty()
    {
        return $this->linkedToThirdParty;
    }

    public function setLinkedToThirdParty($linkedToThirdParty): self
    {
        $this->linkedToThirdParty = $linkedToThirdParty;

        return $this;
    }

    /**
     * @return AddressReference|null
     */
    public function getAddressReference(): ?AddressReference
    {
        return $this->addressReference;
    }

    /**
     * @param AddressReference|null $addressReference
     * @return Address
     */
    public function setAddressReference(?AddressReference $addressReference = null): Address
    {
        $this->addressReference = $addressReference;
        return $this;
    }
}

