<?php

/*
 * Copyright (C) 2015 Julien Fastré <julien.fastre@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\MainBundle\Form\Type;

use Chill\MainBundle\Repository\CenterRepository;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\MainBundle\Security\Authorization\AuthorizationHelperInterface;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\CallbackTransformer;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\Form\FormView;
use Symfony\Component\OptionsResolver\OptionsResolver;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Chill\MainBundle\Entity\Center;
use Chill\MainBundle\Form\Type\DataTransformer\CenterTransformer;
use Symfony\Component\Form\Extension\Core\Type\HiddenType;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Security\Core\Security;

/**
 * Pick a center
 *
 * For a given role and, eventually, scopes, show a dropdown (if more than
 * one reachable center) or a HiddenType (if one or zero center).
 *
 *
 */
class PickCenterType extends AbstractType
{
    protected AuthorizationHelperInterface $authorizationHelper;

    protected Security $security;

    protected CenterRepository $centerRepository;

    public function __construct(
        AuthorizationHelperInterface $authorizationHelper,
        Security $security,
        CenterRepository $centerRepository
    ) {
        $this->authorizationHelper = $authorizationHelper;
        $this->security = $security;
        $this->centerRepository = $centerRepository;
    }

    /**
     * return a 'hidden' field if only one center is available.
     *
     * Return a 'choice' field if more than one center is available.
     *
     * @return string
     * @throws \RuntimeException if the user is not associated with any center
     */
    /*
    public function getParent()
    {
        $nbReachableCenters = count($this->reachableCenters);

        if ($nbReachableCenters <= 1) {
            return HiddenType::class;
        } else {
            return EntityType::class;
        }
    }
    */

    /**
     * configure default options, i.e. add choices if user can reach multiple
     * centers.
     *
     * @param OptionsResolver $resolver
     */
    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver
            ->setDefault('class', Center::class)
            ->setRequired('role')
            ->setAllowedTypes('role', [ 'string' ])
            ->setDefault('scopes', [])
            ->setAllowedTypes('scopes', ['iterable'])
            ->setDefault('choice_options', [])
            ;
        /*
            ->setDefault('choices', $this->reachableCenters)
            ->setDefault('placeholder', 'Pick a center')
            ;
        */
    }

    /**
     * add a data transformer if user can reach only one center
     *
     * @param FormBuilderInterface $builder
     * @param array $options
     */
    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $centers = $this->getReachableCenters($options['role'], $options['scopes']);

        if (count($centers) <= 1) {
            $multiple = $options['choice_options']['multiple'] ?? false;
            $builder->add('center', HiddenType::class);
            $builder->get('center')->addModelTransformer(
               new CenterTransformer($this->centerRepository, $multiple)
            );
        } else {
            $builder->add('center', EntityType::class,
                \array_merge(
                    $options['choice_options'],
                    [
                        'class' => Center::class,
                        'choices' => $centers
                    ]
                )
            );
        }

        $builder
            ->addModelTransformer(new CallbackTransformer(
                function($data) {
                    if (NULL === $data) {
                        return ['center' => null];
                    }
                    return ['center' => $data];
                },
                function($data) {
                    return $data['center'];
                }
            ));
    }

    private function getReachableCenters(string $role, iterable $scopes): array
    {
        if (0 < count($scopes)) {
            $centers = [];

            foreach($scopes as $scope) {
                foreach ($this->authorizationHelper
                             ->getReachableCenters($this->security->getUser(), $role, $scope) as $center) {
                    $centers[spl_object_hash($center)] = $center;
                }
            }

            return \array_values($centers);
        } else {
           return $this->authorizationHelper
               ->getReachableCenters($this->security->getUser(), $role);
        }
    }

    public function buildView(FormView $view, FormInterface $form, array $options)
    {
        $view->vars['is_hidden'] = count($this->getReachableCenters($options['role'],
            $options['scopes'])) <= 1;
    }
}
