<?php
/*
 * Copyright (C) 2017 Champs Libres Cooperative <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\MainBundle\Form\Type;

use Chill\MainBundle\Entity\Center;
use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Form\DataMapper\ScopePickerDataMapper;
use Chill\MainBundle\Repository\ScopeRepository;
use Chill\MainBundle\Repository\UserACLAwareRepositoryInterface;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\MainBundle\Security\Authorization\AuthorizationHelperInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\EntityRepository;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\Extension\Core\Type\HiddenType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\Form\FormView;
use Symfony\Component\OptionsResolver\Options;
use Symfony\Component\OptionsResolver\OptionsResolver;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Security\Core\Security;

/**
 * Allow to pick amongst available scope for the current
 * user.
 *
 * options :
 *
 * - `center`: the center of the entity
 * - `role`  : the role of the user
 *
 */
class ScopePickerType extends AbstractType
{
    protected AuthorizationHelperInterface $authorizationHelper;

    /**
     * @var TokenStorageInterface
     */
    protected $tokenStorage;

    /**
     * @var ScopeRepository
     */
    protected $scopeRepository;

    /**
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;

    protected Security $security;

    public function __construct(
        AuthorizationHelperInterface $authorizationHelper,
        TokenStorageInterface $tokenStorage,
        ScopeRepository $scopeRepository,
        Security $security,
        TranslatableStringHelper $translatableStringHelper
    ) {
        $this->authorizationHelper = $authorizationHelper;
        $this->tokenStorage = $tokenStorage;
        $this->scopeRepository = $scopeRepository;
        $this->security = $security;
        $this->translatableStringHelper = $translatableStringHelper;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $items = $this->authorizationHelper->getReachableScopes($this->security->getUser(),
            $options['role'], $options['center']);

        if (1 !== count($items)) {
            $builder->add('scope', EntityType::class, [
                'class' => Scope::class,
                'placeholder' => 'Choose the circle',
                'choice_label' => function (Scope $c) {
                    return $this->translatableStringHelper->localize($c->getName());
                },
                'choices' => $items,
            ]);
            $builder->setDataMapper(new ScopePickerDataMapper());
        } else {
            $builder->add('scope', HiddenType::class, [
                'data' => $items[0]->getId(),
            ]);
            $builder->setDataMapper(new ScopePickerDataMapper($items[0]));
        }
    }

    public function buildView(FormView $view, FormInterface $form, array $options)
    {
        $view->vars = array_replace(
            $view->vars, [
                'hideLabel' => true,
            ]
        );
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver
          // create `center` option
            ->setRequired('center')
            ->setAllowedTypes('center', [Center::class, 'array', 'null'])
          // create ``role` option
            ->setRequired('role')
            ->setAllowedTypes('role', ['string', Role::class]);
    }

    /**
     * @param Center|array|Center[] $center
     * @param string $role
     * @return \Doctrine\ORM\QueryBuilder
     */
    protected function buildAccessibleScopeQuery($center, $role)
    {
        $roles = $this->authorizationHelper->getParentRoles($role);
        $roles[] = $role;
        $centers = $center instanceof Center ? [$center]: $center;

        $qb = $this->scopeRepository->createQueryBuilder('s');
        $qb
          // jointure to center
            ->join('s.roleScopes', 'rs')
            ->join('rs.permissionsGroups', 'pg')
            ->join('pg.groupCenters', 'gc')
          // add center constraint
            ->where($qb->expr()->in('IDENTITY(gc.center)', ':centers'))
            ->setParameter('centers', \array_map(fn(Center $c) => $c->getId(), $centers))
          // role constraints
            ->andWhere($qb->expr()->in('rs.role', ':roles'))
            ->setParameter('roles', $roles)
          // user contraint
          ->andWhere(':user MEMBER OF gc.users')
          ->setParameter('user', $this->tokenStorage->getToken()->getUser());

        return $qb;
    }
}
