<?php

namespace Chill\MainBundle\Search;

class SearchApiQuery
{
    private array $select = [];
    private array $selectParams = [];
    private ?string $selectKey = null;
    private array $selectKeyParams = [];
    private ?string $jsonbMetadata = null;
    private array $jsonbMetadataParams = [];
    private ?string $pertinence = null;
    private array $pertinenceParams = [];
    private ?string $fromClause = null;
    private array $fromClauseParams = [];
    private array $whereClauses = [];
    private array $whereClausesParams = [];

    public function addSelectClause(string $select, array $params = []): self
    {
        $this->select[] = $select;
        $this->selectParams = [...$this->selectParams, ...$params];

        return $this;
    }

    public function resetSelectClause(): self
    {
        $this->select = [];
        $this->selectParams = [];
        $this->selectKey = null;
        $this->selectKeyParams = [];
        $this->jsonbMetadata = null;
        $this->jsonbMetadataParams = [];
        $this->pertinence = null;
        $this->pertinenceParams = [];

        return $this;
    }

    public function getSelectClauses(): array
    {
        return $this->select;
    }

    public function getSelectParams(): array
    {
        return $this->selectParams;
    }

    public function setSelectKey(string $selectKey, array $params = []): self
    {
        $this->selectKey = $selectKey;
        $this->selectKeyParams = $params;

        return $this;
    }

    public function setSelectJsonbMetadata(string $jsonbMetadata, array $params = []): self
    {
        $this->jsonbMetadata = $jsonbMetadata;
        $this->jsonbMetadataParams = $params;

        return $this;
    }

    public function setSelectPertinence(string $pertinence, array $params = []): self
    {
        $this->pertinence = $pertinence;
        $this->pertinenceParams = $params;

        return $this;
    }

    public function setFromClause(string $fromClause, array $params = []): self
    {
        $this->fromClause = $fromClause;
        $this->fromClauseParams = $params;

        return $this;
    }

    public function getFromClause(): string
    {
        return $this->fromClause;
    }

    public function getFromParams(): array
    {
        return $this->fromClauseParams;
    }

    /**
     * Set the where clause and replace all existing ones.
     *
     */
    public function setWhereClauses(string $whereClause, array $params = []): self
    {
        $this->whereClauses = [$whereClause];
        $this->whereClausesParams = $params;

        return $this;
    }

    /**
     * Add a where clause.
     *
     * This will add to previous where clauses with and `AND` join
     *
     * @param string $whereClause
     * @param array $params
     * @return $this
     */
    public function andWhereClause(string $whereClause, array $params = []): self
    {
        $this->whereClauses[] = $whereClause;
        \array_push($this->whereClausesParams, ...$params);

        return $this;
    }

    private function buildSelectParams(bool $count = false): array
    {
        if ($count) {
            return [];
        }

        $args = $this->getSelectParams();

        if (null !== $this->selectKey) {
            $args = [...$args, ...$this->selectKeyParams];
        }
        if (null !== $this->jsonbMetadata) {
            $args = [...$args, ...$this->jsonbMetadataParams];
        }
        if (null !== $this->pertinence) {
            $args = [...$args, ...$this->pertinenceParams];
        }

        return $args;
    }

    private function buildSelectClause(bool $countOnly = false): string
    {
        if ($countOnly) {
            return 'count(*) AS c';
        }

        $selects = $this->getSelectClauses();

        if (null !== $this->selectKey) {
            $selects[] = \strtr("'{key}' AS key", [ '{key}' => $this->selectKey ]);
        }
        if (null !== $this->jsonbMetadata) {
            $selects[] = \strtr('{metadata} AS metadata', [ '{metadata}' => $this->jsonbMetadata]);
        }
        if (null !== $this->pertinence) {
            $selects[] = \strtr('{pertinence} AS pertinence', [ '{pertinence}' => $this->pertinence]);
        }

        return \implode(', ', $selects);
    }

    public function buildQuery(bool $countOnly = false): string
    {
        $isMultiple = count($this->whereClauses);
        $where =
            ($isMultiple ? '(' : '').
            \implode(
                ($isMultiple ? ')' : '').' AND '.($isMultiple ? '(' : '')
            , $this->whereClauses).
            ($isMultiple ? ')' : '')
        ;

        $select = $this->buildSelectClause($countOnly);


        return \strtr("SELECT
                {select}
            FROM {from}
            WHERE {where}
        ", [
            '{select}' => $select,
            '{from}' => $this->fromClause,
            '{where}' => $where,
        ]);
    }


    public function buildParameters(bool $countOnly = false): array
    {
        if (!$countOnly) {
            return [
               ...$this->buildSelectParams($countOnly),
               ...$this->fromClauseParams,
               ...$this->whereClausesParams,
            ];
        } else {
            return [
                ...$this->fromClauseParams,
                ...$this->whereClausesParams,
            ];
        }
    }
}
