<?php

namespace Chill\PersonBundle\Entity\Household;

use Doctrine\ORM\Mapping as ORM;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Entity\Household\Household;
use Chill\PersonBundle\Entity\Household\Position;
use Symfony\Component\Serializer\Annotation as Serializer;
use Symfony\Component\Validator\Constraints as Assert;


/**
 * @ORM\Entity
 * @ORM\Table(
 *      name="chill_person_household_members"
 *  )
 */
class HouseholdMember
{
    /**
     * @ORM\Id
     * @ORM\GeneratedValue
     * @ORM\Column(type="integer")
     * @Serializer\Groups({"read"})
     */
    private $id;

    /**
     * @ORM\ManyToOne(targetEntity=Position::class)
     * @Serializer\Groups({"read"})
     * @Assert\NotNull(groups={"household_memberships_created"})
     */
    private ?Position $position = null;

    /**
     * @ORM\Column(type="date_immutable", nullable=true, options={"default": null})
     * @Serializer\Groups({"read"})
     * @Assert\NotNull(groups={"household_memberships"})
     */
    private ?\DateTimeImmutable $startDate = null;

    /**
     * @ORM\Column(type="date_immutable", nullable= true, options={"default": null})
     * @Serializer\Groups({"read"})
     * @Assert\GreaterThan(
     *      propertyPath="startDate",
     *      message="household_membership.The end date must be after start date",
     *      groups={"household_memberships"}
     * )
     */
    private ?\DateTimeImmutable $endDate = null;

    /**
     * @ORM\Column(type="string", length=255, nullable=true)
     * @Serializer\Groups({"read"})
     */
    private ?string $comment = NULL;

    /**
     * @ORM\Column(type="boolean", name="sharedhousehold")
     */
    private bool $shareHousehold = false;

    /**
     * @ORM\Column(type="boolean", options={"default": false})
     * @Serializer\Groups({"read"})
     */
    private bool $holder = false;

    /**
     *
     * @var Person
     * @ORM\ManyToOne(
     *  targetEntity="\Chill\PersonBundle\Entity\Person"
     * )
     * @Serializer\Groups({"read"})
     * @Assert\Valid(groups={"household_memberships"})
     * @Assert\NotNull(groups={"household_memberships"})
     */
    private ?Person $person = null;

    /**
     *
     * @var Household
     * @ORM\ManyToOne(
     *  targetEntity="\Chill\PersonBundle\Entity\Household\Household"
     * )
     * @Assert\Valid(groups={"household_memberships"})
     * @Assert\NotNull(groups={"household_memberships"})
     */
    private ?Household $household = null;


    public function getId(): ?int
    {
        return $this->id;
    }

    public function getPosition(): ?Position
    {
        return $this->position;
    }

    public function setPosition(Position $position): self
    {
        if ($this->position instanceof Position) {
            throw new \LogicException("The position is already set. You cannot change ".
                "a position of a membership");
        }

        $this->position = $position;
        $this->shareHousehold = $position->getShareHousehold();

        return $this;
    }

    public function getStartDate(): ?\DateTimeImmutable
    {
        return $this->startDate;
    }

    public function setStartDate(\DateTimeImmutable $startDate): self
    {
        $this->startDate = $startDate;

        return $this;
    }

    public function getEndDate(): ?\DateTimeImmutable
    {
        return $this->endDate;
    }

    public function setEndDate(?\DateTimeImmutable $endDate = null): self
    {
        $this->endDate = $endDate;

        return $this;
    }

    public function getComment(): ?string
    {
        return $this->comment;
    }

    public function setComment(?string $comment): self
    {
        $this->comment = $comment;

        return $this;
    }

    /**
     * @Serializer\Groups({"read"})
     */
    public function getShareHousehold(): ?bool
    {
        return $this->shareHousehold;
    }

    public function setShareHousehold(bool $shareHousehold): self
    {
        $this->shareHousehold = $shareHousehold;

        return $this;
    }

    public function getPerson(): ?Person
    {
        return $this->person;
    }

    public function setPerson(?Person $person): self
    {
        if ($this->person instanceof Person) {
            throw new \LogicException("You cannot change person ".
                "on a membership");
        }

        $this->person = $person;
        $this->person->addHouseholdParticipation($this);

        return $this;
    }

    public function getHousehold(): ?Household
    {
        return $this->household;
    }

    public function setHousehold(?Household $household): self
    {
        if ($this->household instanceof Household) {
            throw new \LogicException("You cannot change household ".
                "on a membership");
        }

        $this->household = $household;

        return $this;
    }

    public function setHolder(bool $holder): self
    {
        $this->holder = $holder;

        return $this;
    }

    public function isHolder(): bool
    {
        return $this->holder;
    }

    public function isCurrent(\DateTimeImmutable $at = null): bool
    {
        $at = NULL === $at ? new \DateTimeImmutable('now'): $at;

        return $this->getStartDate() < $at && (
            NULL === $this->getEndDate() || $at < $this->getEndDate()
        );
    }
}
