<?php

/*
 * Chill is a software for social workers
 *
 * Copyright (C) 2014, Champs Libres Cooperative SCRLFS, <http://www.champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\PersonBundle\Form;

use Chill\CustomFieldsBundle\Form\Type\CustomFieldType;
use Chill\MainBundle\Entity\Civility;
use Chill\MainBundle\Form\Type\ChillCollectionType;
use Chill\MainBundle\Form\Type\ChillTextareaType;
use Chill\MainBundle\Form\Type\Select2CountryType;
use Chill\MainBundle\Form\Type\Select2LanguageType;
use Chill\PersonBundle\Config\ConfigPersonAltNamesHelper;
use Chill\PersonBundle\Form\Type\GenderType;
use Chill\PersonBundle\Form\Type\PersonAltNameType;
use Chill\PersonBundle\Form\Type\PersonPhoneType;
use Chill\PersonBundle\Entity\PersonPhone;
use Chill\PersonBundle\Form\Type\Select2MaritalStatusType;
use Symfony\Component\Form\AbstractType;
use Chill\MainBundle\Form\Type\ChillDateType;
use Chill\MainBundle\Form\Type\CommentType;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\QueryBuilder;
use Doctrine\ORM\EntityRepository;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\CallbackTransformer;
use Symfony\Component\Form\Extension\Core\Type\EmailType;
use Symfony\Component\Form\Extension\Core\Type\TelType;
use Symfony\Component\Form\Extension\Core\Type\TextType;
use Symfony\Component\Form\Extension\Core\Type\IntegerType;
use Symfony\Component\Form\Extension\Core\Type\CheckboxType;
use Symfony\Component\Form\Extension\Core\Type\DateType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;

class PersonType extends AbstractType
{
    /**
     * array of configuration for person_fields.
     *
     * Contains whether we should add fields some optional fields (optional per
     * instance)
     *
     * @var string[]
     */
    protected $config = array();

    /**
     *
     * @var ConfigPersonAltNamesHelper
     */
    protected $configAltNamesHelper;

    protected TranslatableStringHelper $translatableStringHelper;

    /**
     *
     * @param string[] $personFieldsConfiguration configuration of visibility of some fields
     */
    public function __construct(
        array $personFieldsConfiguration,
        ConfigPersonAltNamesHelper $configAltNamesHelper,
        TranslatableStringHelper $translatableStringHelper
    ) {
        $this->config = $personFieldsConfiguration;
        $this->configAltNamesHelper = $configAltNamesHelper;
        $this->translatableStringHelper = $translatableStringHelper;
    }

    /**
     * @param FormBuilderInterface $builder
     * @param array $options
     */
    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $builder
            ->add('firstName')
            ->add('lastName')
            ->add('birthdate', ChillDateType::class, [
                'required' => false,
            ])
            ->add('deathdate', DateType::class, [
                'required' => false,
                'input' => 'datetime_immutable',
                'widget' => 'single_text'
            ])
            ->add('gender', GenderType::class, array(
                'required' => true
            ))
            ->add('genderComment', CommentType::class, array(
                'required' => false
            ))
            ->add('numberOfChildren', IntegerType::class, array(
                'required' => false
            ));

        if ($this->configAltNamesHelper->hasAltNames()) {
            $builder->add('altNames', PersonAltNameType::class, [
                'by_reference' => false
            ]);
        }

        if ($this->config['memo'] === 'visible') {
            $builder
                ->add('memo', ChillTextareaType::class, array('required' => false))
            ;
        }

        if ($this->config['place_of_birth'] === 'visible') {
            $builder->add('placeOfBirth', TextType::class, array(
                'required' => false,
                'attr' => ['style' => 'text-transform: uppercase;'],
            ));

            $builder->get('placeOfBirth')->addModelTransformer(new CallbackTransformer(
                function ($string) {
                    return strtoupper($string);
                },
                function ($string) {
                    return strtoupper($string);
                }
                ));
        }

        if ($this->config['contact_info'] === 'visible') {
            $builder->add('contactInfo', ChillTextareaType::class, array('required' => false));
        }

        if ($this->config['phonenumber'] === 'visible') {
            $builder->add('phonenumber', TelType::class, array(
                'required' => false,
                // 'placeholder' => '+33623124554' //TODO placeholder for phone numbers
            ));
        }

        if ($this->config['mobilenumber'] === 'visible') {
            $builder
                ->add('mobilenumber', TelType::class, array('required' => false))
                ->add('acceptSMS', CheckboxType::class, array(
                    'required' => false
                ));
        }

        $builder->add('otherPhoneNumbers', ChillCollectionType::class, [
            'entry_type' => PersonPhoneType::class,
            'button_add_label' => 'Add new phone',
            'button_remove_label' => 'Remove phone',
            'required' => false,
            'allow_add' => true,
            'allow_delete' => true,
            'by_reference' => false,
            'label' => false,
            'delete_empty' => function(PersonPhone $pp = null) {
                return NULL === $pp || $pp->isEmpty();
            },
            'error_bubbling' => false,
            'empty_collection_explain' => 'No additional phone numbers'
        ]);

        if ($this->config['email'] === 'visible') {
            $builder
                ->add('email', EmailType::class, array('required' => false));
        }

        if ($this->config['acceptEmail'] === 'visible') {
            $builder
                ->add('acceptEmail', CheckboxType::class, array('required' => false));
        }

        if ($this->config['country_of_birth'] === 'visible') {
            $builder->add('countryOfBirth', Select2CountryType::class, array(
                'required' => false
                ));
        }

        if ($this->config['nationality'] === 'visible') {
            $builder->add('nationality', Select2CountryType::class, array(
                'required' => false
                ));
        }

        if ($this->config['spoken_languages'] === 'visible') {
            $builder->add('spokenLanguages', Select2LanguageType::class, array(
                'required' => false,
                'multiple' => true
                ));
        }

        if ($this->config['civility'] === 'visible'){
            $builder
                ->add('civility', EntityType::class, [
                    'label' => 'Civility',
                    'class' => Civility::class,
                    'choice_label' => function (Civility $civility): string {
                        return $this->translatableStringHelper->localize($civility->getName());
                    },
                    'query_builder' => function (EntityRepository $er): QueryBuilder {
                        return $er->createQueryBuilder('c')
                            ->where('c.active = true');
                    },
                    'placeholder' => 'choose civility',
                    'required' => false
                ]);
        }

        if ($this->config['marital_status'] === 'visible'){
            $builder
                ->add('maritalStatus', Select2MaritalStatusType::class, array(
                    'required' => false
                    ))
                ->add('maritalStatusDate', ChillDateType::class, array(
                    'required' => false
                    ))
                ->add('maritalStatusComment', CommentType::class, array(
                    'required' => false
                    ));
        }

        if($options['cFGroup']) {
            $builder
                ->add('cFData', CustomFieldType::class,
                array('attr' => array('class' => 'cf-fields'), 'group' => $options['cFGroup']))
            ;
        }

    }

    /**
     * @param OptionsResolverInterface $resolver
     */
    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setDefaults(array(
            'data_class' => 'Chill\PersonBundle\Entity\Person',
            'validation_groups' => array('general', 'creation'),
        ));

        $resolver->setRequired(array(
            'cFGroup'
        ));

        $resolver->setAllowedTypes(
            'cFGroup', array('null', 'Chill\CustomFieldsBundle\Entity\CustomFieldsGroup')
        );
    }

    /**
     * @return string
     */
    public function getBlockPrefix()
    {
        return 'chill_personbundle_person';
    }
}
