import 'es6-promise/auto';
import { createStore } from 'vuex';
import { fetchScopes } from 'ChillMainAssets/lib/api/scope.js';
import { getAccompanyingCourse,
   patchAccompanyingCourse,
   confirmAccompanyingCourse,
   postParticipation,
   postRequestor,
   postResource,
   postSocialIssue,
   addScope,
   removeScope,
   getReferrersSuggested,
   getUsers,
} from '../api';
import { patchPerson } from "ChillPersonAssets/vuejs/_api/OnTheFly";
import { patchThirdparty } from "ChillThirdPartyAssets/vuejs/_api/OnTheFly";


const debug = process.env.NODE_ENV !== 'production';
const id = window.accompanyingCourseId;

let scopesPromise = fetchScopes();
let accompanyingCoursePromise = getAccompanyingCourse(id);

let initPromise = Promise.all([scopesPromise, accompanyingCoursePromise])
  .then(([scopes, accompanyingCourse]) => new Promise((resolve, reject) => {

      const store = createStore({
         strict: debug,
         modules: {
         },
         state: {
            accompanyingCourse: accompanyingCourse,
            addressContext: {},
            errorMsg: [],
            // all the available scopes
            scopes: scopes,
            // the scopes at start. If the user remove all scopes, we re-add those scopes, by security
            scopesAtStart: accompanyingCourse.scopes.map(scope => scope),
            // the scope states at server side
            scopesAtBackend: accompanyingCourse.scopes.map(scope => scope),
            // the users which are available for referrer
            referrersSuggested: [],
            // all the users available
            users: [],
         },
         getters: {
            isParticipationValid(state) {
               return state.accompanyingCourse.participations.length > 0;
            },
            isSocialIssueValid(state) {
               return state.accompanyingCourse.socialIssues.length > 0;
            },
            isOriginValid(state) {
               return state.accompanyingCourse.origin !== null;
            },
            isLocationValid(state) {
               return state.accompanyingCourse.location !== null;
            },
            isScopeValid(state) {
               //console.log('is scope valid', state.accompanyingCourse.scopes.length > 0);
               return state.accompanyingCourse.scopes.length > 0;
            },
            validationKeys(state, getters) {
               let keys = [];
               if (!getters.isParticipationValid) { keys.push('participation'); }
               if (!getters.isLocationValid) { keys.push('location'); }
               if (!getters.isSocialIssueValid) { keys.push('socialIssue'); }
               if (!getters.isOriginValid) { keys.push('origin'); }
               if (!getters.isScopeValid) { keys.push('scopes'); }
               //console.log('getter keys', keys);
               return keys;
            },
            isValidToBeConfirmed(state, getters) {
               if (getters.validationKeys.length === 0) {
                  return true;
               }
               return false;
            }
         },
         mutations: {
            catchError(state, error) {
               // console.log('### mutation: a new error have been catched and pushed in store !', error);
               state.errorMsg.push(error);
            },
            removeParticipation(state, participation) {
               //console.log('### mutation: remove participation', participation.id);
               state.accompanyingCourse.participations = state.accompanyingCourse.participations.filter(element => element !== participation);
            },
            closeParticipation(state, { participation, payload }) {
               //console.log('### mutation: close item', { participation, payload });
               // find row position and replace by closed participation
               state.accompanyingCourse.participations.splice(
                  state.accompanyingCourse.participations.findIndex(element => element === payload), 1, participation
               );
            },
            addParticipation(state, participation) {
               //console.log('### mutation: add participation', participation);
               state.accompanyingCourse.participations.push(participation);
            },
            removeRequestor(state) {
               //console.log('### mutation: removeRequestor');
               state.accompanyingCourse.requestor = null;
            },
            addRequestor(state, requestor) {
               //console.log('### mutation: addRequestor', requestor);
               state.accompanyingCourse.requestor = requestor;
            },
            requestorIsAnonymous(state, value) {
               //console.log('### mutation: requestorIsAnonymous', value);
               state.accompanyingCourse.requestorAnonymous = value;
            },
            removeResource(state, resource) {
               //console.log('### mutation: removeResource', resource);
               state.accompanyingCourse.resources = state.accompanyingCourse.resources.filter(element => element !== resource);
            },
            addResource(state, resource) {
               //console.log('### mutation: addResource', resource);
               state.accompanyingCourse.resources.push(resource);
            },
            updatePerson(state, payload) {
               console.log('### mutation: updatePerson', payload);
               let i = null;
               switch (payload.target) {
                  case 'participation':
                     i = state.accompanyingCourse.participations.findIndex(e => e.person.id === payload.person.id );
                     state.accompanyingCourse.participations[i].person = payload.person;
                     break;
                  case 'requestor':
                     state.accompanyingCourse.requestor = payload.person;
                     break;
                  case 'resource':
                     i = state.accompanyingCourse.resources.findIndex(e => e.resource.id === payload.person.id );
                     state.accompanyingCourse.resources[i].resource = payload.person;
                     break;
               }
            },
            updateThirdparty(state, payload) {
               console.log('### mutation: updateThirdparty', payload);
               let i = null;
               switch (payload.target) {
                  case 'requestor':
                     state.accompanyingCourse.requestor = payload.thirdparty;
                     break;
                  case 'resource':
                     i = state.accompanyingCourse.resources.findIndex(e => e.resource.id === payload.thirdparty.id );
                     state.accompanyingCourse.resources[i].resource = payload.thirdparty;
                     break;
               }
            },
            toggleIntensity(state, value) {
               state.accompanyingCourse.intensity = value;
            },
            toggleEmergency(state, value) {
               //console.log('### mutation: toggleEmergency');
               state.accompanyingCourse.emergency = value;
            },
            toggleConfidential(state, value) {
               //console.log('### mutation: toggleConfidential');
               state.accompanyingCourse.confidential = value;
            },
            postFirstComment(state, comment) {
               //console.log('### mutation: postFirstComment', comment);
               state.accompanyingCourse.initialComment = comment;
            },
            updateSocialIssues(state, value) {
               console.log('updateSocialIssues', value);
               state.accompanyingCourse.socialIssues = value;
            },
            refreshSocialIssues(state, issues) {
               console.log('refreshSocialIssues', issues);
               state.accompanyingCourse.socialIssues = issues;
            },
            updateOrigin(state, value) {
               //console.log('value', value);
               state.accompanyingCourse.origin = value;
            },
            updateReferrer(state, value) {
               //console.log('value', value);
               state.accompanyingCourse.user = value;
            },
            setReferrersSuggested(state, users) {
               state.referrersSuggested = users.map(u => {
                  if (state.accompanyingCourse.user !== null) {
                     if (state.accompanyingCourse.user.id === u.id) {
                        return state.accompanyingCourse.user;
                     }
                  }
                  return u;
               });
            },
            confirmAccompanyingCourse(state, response) {
               //console.log('### mutation: confirmAccompanyingCourse: response', response);
               state.accompanyingCourse.step = response.step;
            },
            setAddressContext(state, context) {
               //console.log('define location context');
               state.addressContext = context;
            },
            setUsers(state, users) {
               state.users = users.map(u => {
                  if (state.accompanyingCourse.user !== null) {
                     if (state.accompanyingCourse.user.id === u.id) {
                        return state.accompanyingCourse.user;
                     }
                  }
                  return u;
               });
            },
            updateLocation(state, r) {
               //console.log('### mutation: set location attributes', r);
               state.accompanyingCourse.locationStatus = r.locationStatus;
               if (r.locationStatus !== 'person') {
                  state.addressContext.addressId = r.location.address_id;
                  //console.log('mutation: update context addressId', state.addressContext.addressId);
               }
               state.accompanyingCourse.location = r.location;
               state.accompanyingCourse.personLocation = r.personLocation;
            },
            setEditContextTrue(state, payload) {
               //console.log('### mutation: set edit context true with addressId', payload.addressId);
               state.addressContext.edit = true;
               state.addressContext.addressId = payload.addressId;
            },
            setScopes(state, scopes) {
               state.accompanyingCourse.scopes = scopes;
            },
            addScopeAtBackend(state, scope) {
               let scopeIds = state.scopesAtBackend.map(s => s.id);
               if (!scopeIds.includes(scope.id)) {
                  state.scopesAtBackend.push(scope);
               }
            },
            removeScopeAtBackend(state, scope){
               let scopeIds = state.scopesAtBackend.map(s => s.id);
               if (scopeIds.includes(scope.id)) {
                  state.scopesAtBackend = state.scopesAtBackend.filter(s => s.id !== scope.id);
               }
            }
         },
         actions: {
            removeParticipation({ commit }, payload) {
               commit('removeParticipation', payload);
               // fetch DELETE request...
            },
            closeParticipation({ commit }, payload) {
               //console.log('## action: fetch delete participation: payload', payload);
               postParticipation(id, payload.person, 'DELETE')
               .then(participation => new Promise((resolve, reject) => {
                  commit('closeParticipation', { participation, payload });
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            addParticipation({ commit }, payload) {
               //console.log('## action: fetch post participation (select item): payload', payload);
               postParticipation(id, payload.result, 'POST')
               .then(participation => new Promise((resolve, reject) => {
                  commit('addParticipation', participation);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            removeRequestor({ commit, dispatch }) {
               //console.log('## action: fetch delete requestor');
               postRequestor(id, null, 'DELETE')
               .then(requestor => new Promise((resolve, reject) => {
                  commit('removeRequestor');
                  dispatch('requestorIsAnonymous', false);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            addRequestor({ commit }, payload) {
               //console.log('## action: fetch post requestor: payload', payload);
               postRequestor(id, payload.result, 'POST')
               .then(requestor => new Promise((resolve, reject) => {
                  commit('addRequestor', requestor);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            requestorIsAnonymous({ commit }, payload) {
               //console.log('## action: fetch patch AccompanyingCourse: payload', payload);
               patchAccompanyingCourse(id, { type: "accompanying_period", requestorAnonymous: payload })
               .then(course => new Promise((resolve, reject) => {
                  commit('requestorIsAnonymous', course.requestorAnonymous)
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            removeResource({ commit }, payload) {
               //console.log('## action: fetch postResource: payload', payload);
               postResource(id, payload, 'DELETE')
               .then(resource => new Promise((resolve, reject) => {
                  commit('removeResource', payload) // mieux un retour de l'objet !
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            addResource({ commit }, payload) {
               //console.log('## action: fetch postResource: payload', payload);
               postResource(id, payload.result, 'POST')
               .then(resource => new Promise((resolve, reject) => {
                  commit('addResource', resource)
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            patchOnTheFly({ commit }, payload) {
               // TODO should be into the dedicated component, no ? JF
               console.log('## action: patch OnTheFly', payload);
               let body = { type: payload.type };
               if (payload.type === 'person') {
                  body.firstName = payload.data.firstName;
                  body.lastName = payload.data.lastName;
                  if (payload.data.birthdate !== null) { body.birthdate = payload.data.birthdate; }
                  body.phonenumber = payload.data.phonenumber;
                  body.mobilenumber = payload.data.mobilenumber;
                  body.gender = payload.data.gender;
                  console.log('id', payload.data.id, 'and body', body);
                  patchPerson(payload.data.id, body)
                      .then(person => new Promise((resolve, reject) => {
                         console.log('patch person', person);
                         commit('updatePerson', { target: payload.target, person: person });
                         resolve();
                      }));
               }
               else if (payload.type === 'thirdparty') {
                  body.name = payload.data.text;
                  body.email = payload.data.email;
                  body.telephone = payload.data.phonenumber;
                  body.address = { id: payload.data.address.address_id };
                  console.log('id', payload.data.id, 'and body', body);
                  patchThirdparty(payload.data.id, body)
                      .then(thirdparty => new Promise((resolve, reject) => {
                          console.log('patch thirdparty', thirdparty);
                          commit('updateThirdparty', { target: payload.target, thirdparty: thirdparty });
                          resolve();
                  }));
               }
            },
            toggleIntensity({ commit }, payload) {
               //console.log(payload);
               patchAccompanyingCourse(id, { type: "accompanying_period", intensity: payload })
               .then(course => new Promise((resolve, reject) => {
                  commit('toggleIntensity', course.intensity);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            toggleEmergency({ commit }, payload) {
               patchAccompanyingCourse(id, { type: "accompanying_period", emergency: payload })
                  .then(course => new Promise((resolve, reject) => {
                     commit('toggleEmergency', course.emergency);
                     return dispatch('setRefererresAvailable');
                  }))
                 .then(() => Promise.resolve())
                 .catch((error) => { commit('catchError', error) });
            },
            toggleConfidential({ commit }, payload) {
               patchAccompanyingCourse(id, { type: "accompanying_period", confidential: payload })
               .then(course => new Promise((resolve, reject) => {
                  commit('toggleConfidential', course.confidential);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            /**
             * Handle the checked/unchecked scopes
             *
             * When the user set the scopes in a invalid situation (when no scopes are cheched), this
             * method will internally re-add the scopes as they were originally when the page was loaded, but
             * this does not appears for the user (they remains unchecked). When the user re-add a scope, the
             * scope is back in a valid state, and the store synchronize with the new state (all the original scopes
             * are removed if necessary, and the new checked scopes is backed).
             *
             * So, for instance:
             *
             * at load:
             *
             * [x] scope A (at backend: [x])
             * [x] scope B (at backend: [x])
             * [ ] scope C (at backend: [ ])
             *
             * The user uncheck scope A:
             *
             * [ ] scope A (at backend: [ ] as soon as the operation finish)
             * [x] scope B (at backend: [x])
             * [ ] scope C (at backend: [ ])
             *
             * The user uncheck scope B. The state is invalid (no scope checked), so we go back to initial state when
             * the page loaded):
             *
             * [ ] scope A (at backend: [x] as soon as the operation finish)
             * [ ] scope B (at backend: [x] as soon as the operation finish)
             * [ ] scope C (at backend: [ ])
             *
             * The user check scope C. The scopes are back to valid state. So we go back to synchronization with UI and
             * backend):
             *
             * [ ] scope A (at backend: [ ] as soon as the operation finish)
             * [ ] scope B (at backend: [ ] as soon as the operation finish)
             * [x] scope C (at backend: [x] as soon as the operation finish)
             *
             * **Warning** There is a problem if the user check/uncheck faster than the backend is synchronized.
             *
             * @param commit
             * @param state
             * @param dispatch
             * @param scopes
             * @returns Promise
             */
            setScopes({ commit, state, dispatch }, scopes) {
               let currentServerScopesIds = state.scopesAtBackend.map(scope => scope.id);
               let checkedScopesIds = scopes.map(scope => scope.id);
               let removedScopesIds = currentServerScopesIds.filter(id => !checkedScopesIds.includes(id));
               let addedScopesIds = checkedScopesIds.filter(id => !currentServerScopesIds.includes(id));
               let lengthAfterOperation = currentServerScopesIds.length + addedScopesIds.length
                 - removedScopesIds.length;

               if (lengthAfterOperation > 0 || (lengthAfterOperation === 0 && state.scopesAtStart.length === 0) ) {
                  return dispatch('updateScopes', {
                     addedScopesIds, removedScopesIds
                  }).then(() => {
                     // warning: when the operation of dispatch are too slow, the user may check / uncheck before
                     // the end of the synchronisation with the server (done by dispatch operation). Then, it leads to
                     // check/uncheck in the UI. I do not know of to avoid it.
                     commit('setScopes', scopes);
                     return Promise.resolve();
                  }).then(() => {
                     return dispatch('fetchReferrersSuggested');
                  });
               } else {
                  return dispatch('setScopes', state.scopesAtStart).then(() => {
                     commit('setScopes', scopes);
                     return Promise.resolve();
                  }).then(() => {
                     return dispatch('fetchReferrersSuggested');
                  });
               }
            },
            /**
             * Internal function for the store to effectively update scopes.
             *
             * Return a promise which resolves when all update operation are
             * successful and finished.
             *
             * @param state
             * @param commit
             * @param addedScopesIds
             * @param removedScopesIds
             * @return Promise
             */
            updateScopes({ state, commit }, { addedScopesIds, removedScopesIds }) {
               let promises = [];
               state.scopes.forEach(scope => {
                  if (addedScopesIds.includes(scope.id)) {
                     promises.push(addScope(state.accompanyingCourse.id, scope).then(() => {
                        commit('addScopeAtBackend', scope);
                        return Promise.resolve();
                     }));
                  }
                  if (removedScopesIds.includes(scope.id)) {
                     promises.push(removeScope(state.accompanyingCourse.id, scope).then(() => {
                        commit('removeScopeAtBackend', scope);
                        return Promise.resolve();
                     }));
                  }
               });

               return Promise.all(promises);
            },
            postFirstComment({ commit }, payload) {
               //console.log('## action: postFirstComment: payload', payload);
               patchAccompanyingCourse(id, { type: "accompanying_period", initialComment: payload })
               .then(course => new Promise((resolve, reject) => {
                  commit('postFirstComment', course.initialComment);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            updateSocialIssues({ state, commit, dispatch }, { payload, body, method }) {
               //console.log('## action: payload', { payload, body, method });
               postSocialIssue(id, body, method)
               .then(response => new Promise((resolve, reject) => {
                  //console.log('response', response);
                  commit('updateSocialIssues', payload);
                  resolve();
               })).then(() => {
                  return getAccompanyingCourse(state.accompanyingCourse.id);
               }).then(accompanying_course => {
                  commit('refreshSocialIssues', accompanying_course.socialIssues);
                  dispatch('fetchReferrersSuggested');
               })
                   .catch((error) => { commit('catchError', error) });
            },
            updateOrigin({ commit }, payload) {
               patchAccompanyingCourse(id, { type: "accompanying_period", origin: { id: payload.id, type: payload.type } })
               .then(course => new Promise((resolve, reject) => {
                  commit('updateOrigin', course.origin);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            updateReferrer({ commit }, payload) {
               patchAccompanyingCourse(id, { type: "accompanying_period", user: { id: payload.id, type: payload.type } })
               .then(course => new Promise((resolve, reject) => {
                  commit('updateReferrer', course.user);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            },
            async fetchReferrersSuggested({ state, commit}) {
               let users = await getReferrersSuggested(state.accompanyingCourse);
               commit('setReferrersSuggested', users);
               if (
                 null === state.accompanyingCourse.user
                 && !state.accompanyingCourse.confidential
                 && !state.accompanyingCourse.step === 'DRAFT'
                 && users.length === 1
               ) {
                  // set the user if unique
                  commit('updateReferrer', users[0]);
               }
            },
            async fetchUsers({commit}) {
               let users = await getUsers();
               commit('setUsers', users);
            },
            updateLocation({ commit, dispatch }, payload) {
               //console.log('## action: updateLocation', payload.locationStatusTo);
               let body = { 'type': payload.target, 'id': payload.targetId };
               let location = {};
               if (payload.locationStatusTo === 'person') { // patch for person address (don't remove addressLocation)
                  location = { 'personLocation': { 'type': 'person', 'id': payload.personId }};
               }
               else if (payload.locationStatusTo === 'address') { // patch for temporary address
                  location = { 'personLocation': null, 'addressLocation': { 'id': payload.addressId }};
               }
               else { // patch to remove person address
                  location = { 'personLocation': null };
               }
               Object.assign(body, location);
               patchAccompanyingCourse(payload.targetId, body)
                  .then(accompanyingCourse => new Promise((resolve, reject) => {
                     commit('updateLocation', {
                        location: accompanyingCourse.location,
                        locationStatus: accompanyingCourse.locationStatus,
                        personLocation: accompanyingCourse.personLocation
                     });
                     resolve();
                     dispatch('fetchReferrersSuggested');
                  })).catch((error) => { commit('catchError', error) });
            },
            confirmAccompanyingCourse({ commit }) {
               //console.log('## action: confirmAccompanyingCourse');
               confirmAccompanyingCourse(id)
               .then(response => new Promise((resolve, reject) => {
                  window.location.replace(`/fr/parcours/${id}`);
                  //console.log('fetch resolve');
                  commit('confirmAccompanyingCourse', response);
                  resolve();
               })).catch((error) => { commit('catchError', error) });
            }
         }
      });

      store.dispatch('fetchReferrersSuggested');
      store.dispatch('fetchUsers');
      resolve(store);
   }));

export { initPromise };
