import { createStore } from 'vuex';
import { datetimeToISO, ISOToDatetime, intervalDaysToISO, intervalISOToDays } from 'ChillMainAssets/chill/js/date.js';
import { findSocialActionsBySocialIssue } from 'ChillPersonAssets/vuejs/_api/SocialWorkSocialAction.js';
import { create } from 'ChillPersonAssets/vuejs/_api/AccompanyingCourseWork.js';

const debug = process.env.NODE_ENV !== 'production';
const evalFQDN = encodeURIComponent("Chill\\PersonBundle\\Entity\\AccompanyingPeriod\\AccompanyingPeriodWorkEvaluation");

const store = createStore({
  strict: debug,
  state: {
    work: window.accompanyingCourseWork,
    startDate: window.accompanyingCourseWork.startDate !== null ?
        ISOToDatetime(window.accompanyingCourseWork.startDate.datetime) : null,
    endDate: window.accompanyingCourseWork.endDate !== null ?
        ISOToDatetime(window.accompanyingCourseWork.endDate.datetime) : null,
    note: window.accompanyingCourseWork.note,
    goalsPicked: window.accompanyingCourseWork.goals,
    goalsForAction: [],
    resultsPicked: window.accompanyingCourseWork.results,
    resultsForAction: [],
    resultsForGoal: [],
    evaluationsPicked: window.accompanyingCourseWork.accompanyingPeriodWorkEvaluations.map((e, index) => {
      var k = Object.assign(e, {
        key: index,
        editEvaluation: false,
        startDate: e.startDate !== null ? ISOToDatetime(e.startDate.datetime) : null,
        endDate: e.endDate !== null ? ISOToDatetime(e.endDate.datetime) : null,
        maxDate: e.maxDate !== null ? ISOToDatetime(e.maxDate.datetime) : null,
        warningInterval: e.warningInterval !== null ? intervalISOToDays(e.warningInterval) : null,
      });

      return k;
    }),
    evaluationsForAction: [],
    templatesAvailableForEvaluation: [],
    personsPicked: window.accompanyingCourseWork.persons,
    personsReachables: window.accompanyingCourseWork.accompanyingPeriod.participations.filter(p => p.endDate == null)
      .map(p => p.person),
    handlingThirdParty: window.accompanyingCourseWork.handlingThierParty,
    thirdParties: window.accompanyingCourseWork.thirdParties,
    isPosting: false,
    errors: [],
  },
  getters: {
    socialAction(state) {
      return state.work.socialAction;
    },
    hasResultsForAction(state) {
      return state.resultsForAction.length > 0;
    },
    resultsForGoal: (state) => (goal) => {
      let founds = state.resultsForGoal.filter(r => r.goalId === goal.id);

      return founds === undefined ? [] : founds;
    },
    resultsPickedForGoal: (state) => (goal) => {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      return found === undefined ? [] : found.results;
    },
    hasHandlingThirdParty(state) {
      return state.handlingThirdParty !== null;
    },
    hasThirdParties(state) {
      return state.thirdParties.length > 0;
    },
    getTemplatesAvailaibleForEvaluation(state) {
      return state.templatesAvailableForEvaluation;
    },
    buildPayload(state) {
      return {
        type: 'accompanying_period_work',
        id: state.work.id,
        startDate: state.startDate === null ? null : {
          datetime: datetimeToISO(state.startDate)
        },
        endDate: state.endDate === null ? null : {
          datetime: datetimeToISO(state.endDate)
        },
        note: state.note,
        persons: state.personsPicked.map(p => ({id: p.id, type: p.type})),
        handlingThierParty: state.handlingThirdParty === null ? null : {
          id: state.handlingThirdParty.id,
          type: state.handlingThirdParty.type
        },
        results: state.resultsPicked.map(r => ({id: r.id, type: r.type})),
        thirdParties: state.thirdParties.map(t => ({id: t.id, type: t.type})),
        goals: state.goalsPicked.map(g => {
          let o = {
            type: g.type,
            note: g.note,
            goal: {
              type: g.goal.type,
              id: g.goal.id,
            },
            results: g.results.map(r => ({id: r.id, type: r.type})),
          };
          if (g.id !== undefined) {
            o.id = g.id;
          }
          return o;
        }),
        accompanyingPeriodWorkEvaluations: state.evaluationsPicked.map(e => {
          let o = {
            type: e.type,
            key: e.key,
            evaluation: {
              id: e.evaluation.id,
              type: e.evaluation.type
            },
            startDate: e.startDate !== null ? { datetime: datetimeToISO(e.startDate) } : null,
            endDate: e.endDate !== null ? { datetime: datetimeToISO(e.endDate) } : null,
            maxDate: e.maxDate !== null ? { datetime: datetimeToISO(e.maxDate) } : null,
            warningInterval: intervalDaysToISO(e.warningInterval),
            comment: e.comment,
          };
          if (e.id !== undefined) {
            o.id = e.id;
          }

          return o;
        })
      };
    }
  },
  mutations: {
    setStartDate(state, date) {
      state.startDate = date;
    },
    setEndDate(state, date) {
      state.endDate = date;
    },
    setResultsForAction(state, results) {
      state.resultsForAction = results;
    },
    setResultsForGoal(state, { goal, results }) {
      state.goalsForAction.push(goal);
      for (let i in results) {
        let r = results[i];
        r.goalId = goal.id;
        state.resultsForGoal.push(r);
      }
    },
    setEvaluationsForAction(state, results) {
      state.evaluationsForAction = results;
    },
    addResultPicked(state, result) {
      state.resultsPicked.push(result);
    },
    removeResultPicked(state, result) {
      state.resultsPicked = state.resultsPicked.filter(r => r.id !== result.id);
    },
    addGoal(state, goal) {
      let g = {
        type: "accompanying_period_work_goal",
        goal: goal,
        note: '',
        results: []
      }
      state.goalsPicked.push(g);
    },
    removeGoal(state, goal) {
      state.goalsPicked = state.goalsPicked.filter(g => g.id !== goal.id);
    },
    addResultForGoalPicked(state, { goal, result}) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      if (found === undefined) {
        return;
      }

      found.results.push(result);
    },
    removeResultForGoalPicked(state, { goal, result }) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      if (found === undefined) {
        return;
      }

      found.results = found.results.filter(r => r.id !== result.id);
    },
    addEvaluation(state, evaluation) {
      let e = {
        type: "accompanying_period_work_evaluation",
        key: state.evaluationsPicked.length + 1,
        evaluation: evaluation,
        startDate: null,
        endDate:  null,
        maxDate:  null,
        warningInterval: null,
        comment: "",
        editEvaluation: true,
      };
      state.evaluationsPicked.push(e);
    },
    removeEvaluation(state, evaluation) {
      state.evaluationsPicked = state.evaluationsPicked.filter(e => e.key !== evaluation.key);
    },
    setEvaluationStartDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .startDate = date;
    },
    setEvaluationEndDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .endDate = date;
    },
    setEvaluationMaxDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .maxDate = date;
    },
    setEvaluationWarningInterval(state, {key, days}) {
      state.evaluationsPicked.find(e => e.key === key)
          .warningInterval = days;
    },
    setEvaluationComment(state, {key, comment}) {
      state.evaluationsPicked.find(e => e.key === key)
          .comment = comment;
    },
    toggleEvaluationEdit(state, {key}) {
      let evaluation = state.evaluationsPicked.find(e => e.key === key);
      evaluation.editEvaluation = !evaluation.editEvaluation;
    },
    setTemplatesAvailableForEvaluation(state, templates) {
      for (let i in templates) {
        state.templatesAvailableForEvaluation.push(templates[i]);
      }
    },
    setPersonsPickedIds(state, ids) {
      state.personsPicked = state.personsReachables
        .filter(p => ids.includes(p.id))
    },
    setNote(state, note) {
      state.note = note;
    },
    setHandlingThirdParty(state, thirdParty) {
      state.handlingThirdParty = thirdParty;
    },
    addThirdParties(state, thirdParties) {
      // filter to remove existing thirdparties
      let ids = state.thirdParties.map(t => t.id);
      let unexistings = thirdParties.filter(t => !ids.includes(t.id));

      for (let i in unexistings) {
        state.thirdParties.push(unexistings[i]);
      }
    },
    removeThirdParty(state, thirdParty) {
      state.thirdParties = state.thirdParties
        .filter(t => t.id !== thirdParty.id);
    },
    setErrors(state, errors) {
      state.errors = errors;
    },
    setIsPosting(state, st) {
      state.isPosting = st;
    },
  },
  actions: {
    getReachablesGoalsForAction({ getters, commit, dispatch }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/goal/by-social-action/${socialActionId}.json`
      ;
      window
        .fetch(
          url
        ).then( response => {
          if (response.ok) {
            return response.json();
          }
          throw { m: 'Error while retriving goal for social action', s: response.status, b: response.body };
        }).then( data => {
           for (let i in data.results) {
             dispatch('getReachablesResultsForGoal', data.results[i]);
           }
        }).catch( errors => {
            commit('addErrors', errors);
        });
    },
    getReachablesResultsForGoal({ commit }, goal) {
      let
        url = `/api/1.0/person/social-work/result/by-goal/${goal.id}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for goal', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setResultsForGoal', { goal, results: data.results });
        });
    },
    getReachablesResultsForAction({ getters, commit }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/result/by-social-action/${socialActionId}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for social action', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setResultsForAction', data.results);
        });
    },
    getReachablesEvaluationsForAction({ getters, commit }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/evaluation/by-social-action/${socialActionId}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }
          throw { m: 'Error while retriving evaluations for social action', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setEvaluationsForAction', data.results);
        });
    },
    getReachableTemplatesForEvaluation({commit}) {
      const
          url = `/fr/doc/gen/templates/for/${evalFQDN}`
      ;
      window.fetch(url).then(r => {
        if (r.ok) {
          return r.json();
        }
        throw new Error("not possible to load templates for evaluations")
      }).then(data => {
        commit('setTemplatesAvailableForEvaluation', data.results);
      }).catch(e => {
        console.error(e);
      })
    },
    generateDocument({ dispatch }, {key, templateId}) {
      const callback = function(data) {
        // get the evaluation id from the data
        const
            evaluationId = data.accompanyingPeriodWorkEvaluations.find(e => e.key === key).id,
            returnPath = encodeURIComponent(window.location.pathname + window.location.search + window.location.hash),
            url = `/fr/doc/gen/generate/from/${templateId}/for/${evalFQDN}/${evaluationId}?returnPath=${returnPath}`
        ;
        //http://localhost:8001/fr/doc/gen/generate/from/12/for/Chill%5CPersonBundle%5CEntity%5CAccompanyingPeriod%5CAccompanyingPeriodWorkEvaluation/41

        console.log('I will generate your doc at', url);
        window.location.assign(url);
      };

      dispatch('submit', callback);
    },
    submit({ getters, state, commit }, callback) {
      let
        payload = getters.buildPayload,
        url = `/api/1.0/person/accompanying-course/work/${state.work.id}.json`,
        errors = []
      ;

      commit('setIsPosting', true);

      window.fetch(url, {
        method: 'PUT',
        headers: {
          'Content-Type': 'application/json'
        },
        body: JSON.stringify(payload)
      }).then(response => {
        if (response.ok || response.status === 422) {
          return response.json().then(data => ({data, status: response.status}));
        }

        throw new Error(response.status);
      }).then(({data, status}) => {
        if (status === 422) {
          for (let i in data.violations) {
            errors.push(data.violations[i].title);
          }
          commit('setErrors', errors);
          commit('setIsPosting', false);
        } else if (typeof(callback) !== 'undefined') {
          callback(data);
        } else {
          console.info('nothing to do here, bye bye');
          window.location.assign(`/fr/person/accompanying-period/${state.work.accompanyingPeriod.id}/work`);
        }
      }).catch(e => {
        commit('setErrors', [
          'Erreur serveur ou réseau: veuillez ré-essayer. Code erreur: ' + e
        ]);
        commit('setIsPosting', false);
      });
    },
    initAsync({ dispatch }) {
      dispatch('getReachablesResultsForAction');
      dispatch('getReachablesGoalsForAction');
      dispatch('getReachablesEvaluationsForAction');
      dispatch('getReachableTemplatesForEvaluation');
    },
  }
});

store.dispatch('initAsync');

export { store };
