<?php

namespace Chill\PersonBundle\Validator\Constraints\Household;

use Chill\PersonBundle\Entity\Person;
use Chill\MainBundle\Util\DateRangeCovering;
use Chill\PersonBundle\Templating\Entity\PersonRender;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;
use Symfony\Component\Validator\Exception\UnexpectedTypeException;
use Symfony\Component\Validator\Exception\UnexpectedValueException;

/**
 * Validate that a person does not belong to two household at 
 * the same time
 */
class HouseholdMembershipSequentialValidator extends ConstraintValidator
{
    private PersonRender $render;

    public function __construct(PersonRender $render)
    {
        $this->render = $render;
    }

    public function validate($person, Constraint $constraint)
    {
        if (!$person instanceof Person) {
            throw new UnexpectedTypeException($constraint, Person::class);
        }

        $participations = $person->getHouseholdParticipationsShareHousehold();

        if ($participations->count() === 0) {
            return;
        }

        $covers = new DateRangeCovering(1, $participations->first()
            ->getStartDate()->getTimezone());

        foreach ($participations as $k => $p) {
            $covers->add($p->getStartDate(), $p->getEndDate(), $k);
        }

        $covers->compute();

        if ($covers->hasIntersections()) {
            foreach ($covers->getIntersections() as list($start, $end, $metadata)) {
                $participation = $participations[$metadata[0]];
                $nbHousehold = count($metadata);

                $this->context
                    ->buildViolation($constraint->message)
                    ->setParameters([
                        '%person_name%' => $this->render->renderString(
                            $participation->getPerson(), []
                        ),
                        // TODO when date is correctly i18n, fix this
                        '%from%' => $start->format('d-m-Y'),
                        '%nbHousehold%' => $nbHousehold,

                    ])
                    ->addViolation()
                    ;
            }
        }
    }
}
