<?php

declare(strict_types=1);

namespace Chill\TaskBundle\Security\Authorization;

use Chill\EventBundle\Entity\Event;
use Chill\MainBundle\Entity\Center;
use Chill\MainBundle\Security\Authorization\AbstractChillVoter;
use Chill\MainBundle\Security\Authorization\AuthorizationHelperInterface;
use Chill\MainBundle\Security\Authorization\VoterHelperFactoryInterface;
use Chill\MainBundle\Security\Authorization\VoterHelperInterface;
use Chill\MainBundle\Security\Resolver\CenterResolverDispatcher;
use Chill\TaskBundle\Entity\AbstractTask;
use Symfony\Component\Security\Core\Authorization\AccessDecisionManagerInterface;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use Psr\Log\LoggerInterface;
use Chill\MainBundle\Security\ProvideRoleHierarchyInterface;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Chill\MainBundle\Entity\User;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Chill\TaskBundle\Security\Authorization\AuthorizationEvent;

final class TaskVoter extends AbstractChillVoter implements ProvideRoleHierarchyInterface
{
    public const CREATE_COURSE = 'CHILL_TASK_TASK_CREATE_FOR_COURSE';
    public const CREATE_PERSON = 'CHILL_TASK_TASK_CREATE_FOR_PERSON';
    public const DELETE   = 'CHILL_TASK_TASK_DELETE';
    public const SHOW   = 'CHILL_TASK_TASK_SHOW';
    public const UPDATE = 'CHILL_TASK_TASK_UPDATE';

    public const ROLES = [
        self::CREATE_COURSE,
        self::CREATE_PERSON,
        self::DELETE,
        self::SHOW,
        self::UPDATE,
    ];

    private AccessDecisionManagerInterface $accessDecisionManager;

    private LoggerInterface $logger;

    private EventDispatcherInterface $eventDispatcher;

    private VoterHelperInterface $voter;

    public function __construct(
        AccessDecisionManagerInterface $accessDecisionManager,
        EventDispatcherInterface $eventDispatcher,
        LoggerInterface $logger,
        VoterHelperFactoryInterface $voterFactory
    ) {
        $this->accessDecisionManager = $accessDecisionManager;
        $this->eventDispatcher = $eventDispatcher;
        $this->logger = $logger;

        $this->voter = $voterFactory
            ->generate(AbstractTask::class)
            ->addCheckFor(AbstractTask::class, self::ROLES)
            ->addCheckFor(Person::class, [self::SHOW, self::CREATE_PERSON])
            ->addCheckFor(AccompanyingPeriod::class, [self::SHOW, self::CREATE_COURSE])
            ->addCheckFor(null, [self::SHOW])
            ->build();
    }

    public function supports($attribute, $subject)
    {
        return $this->voter->supports($attribute, $subject);
    }

    protected function voteOnAttribute($attribute, $subject, TokenInterface $token)
    {
        $this->logger->debug(sprintf("Voting from %s class", self::class));

        if (!$token->getUser() instanceof User) {
            return false;
        }

        $event = new AuthorizationEvent($subject, $attribute, $token);

        $this->eventDispatcher->dispatch(AuthorizationEvent::VOTE, $event);

        if ($event->hasVote()) {
            $this->logger->debug("The TaskVoter is overriding by "
                .AuthorizationEvent::VOTE, [
                    'vote' => $event->getVote(),
                    'task_id' => $subject->getId()
                ]);

            return $event->getVote();
        }

        // do pre-flight check, relying on other decision manager
        // those pre-flight check concern associated entities
        if ($subject instanceof AbstractTask) {
            // a user can always see his own tasks
            if ($subject->getAssignee() === $token->getUser()) {
                return true;
            }

            if (NULL !== $person = $subject->getPerson()) {
                if (!$this->accessDecisionManager->decide($token, [PersonVoter::SEE], $person)) {
                    return false;
                }
            } elseif (NULL !== $period = $subject->getCourse()) {
                if (!$this->accessDecisionManager->decide($token, [AccompanyingPeriodVoter::SEE], $period)) {
                    return false;
                }
            }
        }

        // do regular check.
        return $this->voter->voteOnAttribute($attribute, $subject, $token);
    }

    public function getRoles()
    {
        return self::ROLES;
    }

    public function getRolesWithHierarchy(): array
    {
        return [
            'Task' => self::ROLES
        ];
    }

    public function getRolesWithoutScope()
    {
        return [];
    }
}
